use crate::prelude::*;
use serde::de::DeserializeOwned;
use sled::{Db, IVec, Tree};

#[derive(Clone)]
pub struct Data {
    db: Db,
}

impl Data {
    pub fn try_new() -> AnyResult<Self> {
        Ok(Self {
            db: sled::open("data/lyrs")?,
        })
    }

    pub fn delete<K, V>(&self, tree_name: &str, key: K) -> AnyResult<Option<V>>
    where
        K: AsRef<[u8]>,
        V: DeserializeOwned,
    {
        match self.db.open_tree(tree_name)?.remove(key.as_ref())? {
            Some(v) => Ok(Some(bincode::deserialize::<V>(&v)?)),
            None => Ok(None),
        }
    }

    pub fn get<K, V>(&self, tree_name: &str, key: K) -> AnyResult<Option<V>>
    where
        K: AsRef<[u8]>,
        V: DeserializeOwned,
    {
        match self.db.open_tree(tree_name)?.get(key.as_ref())? {
            Some(v) => Ok(Some(bincode::deserialize::<V>(&v)?)),
            None => Ok(None),
        }
    }

    pub fn insert<K: AsRef<[u8]>, V: Into<IVec>>(
        &self,
        tree_name: &str,
        key: K,
        value: V,
    ) -> AnyResult<Option<IVec>> {
        Ok(self.db.open_tree(tree_name)?.insert(key, value.into())?)
    }

    pub fn all<'de, K, V>(
        &self,
        tree_name: &str,
    ) -> AnyResult<impl Iterator<Item = AnyResult<(K, V)>>>
    where
        V: DeserializeOwned,
        K: From<IVec>,
    {
        Ok(self
            .db
            .open_tree(tree_name)?
            .scan_prefix([])
            .map(|r| match r {
                Ok((k, v)) => {
                    let key = K::from(k);
                    match bincode::deserialize::<V>(&v) {
                        Ok(v) => Ok((key, v)),
                        Err(err) => Err(err.into()),
                    }
                }
                Err(err) => Err(err.into()),
            })
            .into_iter())
    }

    pub fn tree(&self, name: &str) -> AnyResult<Tree> {
        Ok(self.db.open_tree(name)?)
    }
}
