use crate::partials::page;
use crate::service::accounts::AuthSession;
use crate::user::User;
use crate::{
    file_watcher::FileWatcher,
    prelude::*,
    service::{accounts, static_files},
    state::State as AppState,
};
use axum::extract::State;
use axum::http::StatusCode;
use axum::response::{Html, IntoResponse};
use axum::{routing::get, Router};
use axum_login::{login_required, AuthManagerLayerBuilder};
use maud::html;
use sled::IVec;
use tower_http::trace::TraceLayer;
use tower_livereload::LiveReloadLayer;
use tower_sessions::SessionManagerLayer;

#[derive(Debug)]
#[allow(dead_code)]
pub struct WebError(Error);

impl From<Error> for WebError {
    fn from(value: Error) -> Self {
        Self(value)
    }
}

impl From<bincode::ErrorKind> for WebError {
    fn from(value: bincode::ErrorKind) -> Self {
        Self(value.into())
    }
}

impl From<Box<bincode::ErrorKind>> for WebError {
    fn from(value: Box<bincode::ErrorKind>) -> Self {
        Self(value.into())
    }
}

impl IntoResponse for WebError {
    fn into_response(self) -> axum::http::Response<axum::body::Body> {
        error!("webserver error: {:?}", self);
        (StatusCode::INTERNAL_SERVER_ERROR, "internal server error").into_response()
    }
}

pub type WebResult<T> = Result<T, WebError>;

pub async fn router(
    state: AppState,
    with_watchers: bool,
) -> AnyResult<(Router, Vec<Option<FileWatcher>>)> {
    let live_reload_layer: Option<LiveReloadLayer> = if with_watchers {
        Some(LiveReloadLayer::new())
    } else {
        None
    };

    let orl = || {
        if let Some(lr) = &live_reload_layer {
            Some(lr.reloader())
        } else {
            None
        }
    };

    let (static_file_service, static_file_watcher) = static_files::router(orl())?;
    let accounts_service = accounts::router(state.clone()).unwrap();

    let session_store = tower_sessions_sled_store::SledStore::new(state.db.tree("session")?);
    let session_layer = SessionManagerLayer::new(session_store);

    let auth_layer = AuthManagerLayerBuilder::new(state.clone(), session_layer).build();

    let mut result = Router::new()
        .route("/dashboard", get(dashboard))
        .route_layer(login_required!(AppState, login_url = "/accounts/login"))
        .route("/", get(index))
        .route("/about", get(about))
        .route("/users", get(users))
        // TODO: admin-only pages
        .nest_service("/accounts", accounts_service)
        .nest_service("/static", static_file_service)
        .layer(TraceLayer::new_for_http())
        .layer(auth_layer)
        .with_state(state.clone());

    if let Some(lr) = live_reload_layer {
        result = result.clone().layer(lr);
    }

    let watchers = vec![static_file_watcher];

    Ok((result, watchers))
}

async fn index(auth_session: Option<AuthSession>) -> Html<String> {
    page(
        "index",
        html! {
            main class="p-2" {
                h1 class="text-2xl" { "Index" }
                p class="mt-2" {
                    "Here, we explain to you why you may like this web application."
                }
            }
        },
        auth_session,
    )
}

async fn about(auth_session: Option<AuthSession>) -> Html<String> {
    page(
        "about",
        html! {
            main class="p-2" {
                h1 class="text-2xl" { "About" }
                p class="mt-2" {
                    "Here, we give a little history on why we made this."
                }
            }
        },
        auth_session,
    )
}

async fn dashboard(auth_session: Option<AuthSession>) -> Html<String> {
    page(
        "dashboard",
        html! {
            div class="p-2" { "Dashboard" }
        },
        auth_session,
    )
}

async fn users(State(state): State<AppState>) -> WebResult<String> {
    let mut s = String::new();
    let mut users = state.db.all::<IVec, User>(User::tree())?;
    while let Some(Ok((_, user))) = users.next() {
        s.push_str(&format!("{}: {:?}", user.username, user.registered_at))
    }
    Ok(s)
}
